﻿using System;
using System.Linq;
using System.Web;
using Klarna.Api;
using Mediachase.Commerce.Orders;
using Mediachase.Commerce.Orders.Exceptions;
using Mediachase.Commerce.Plugins.Payment;
using Mediachase.Commerce.Website;
using EPiServer.Business.Commerce.Payment.Klarna.Orders;
using EPiServer.Logging;

namespace EPiServer.Business.Commerce.Payment.Klarna.KlarnaPartPayment
{
    /// <summary>
    /// Represents Klarna Part payment gateway.
    /// </summary>
    public class KlarnaPaymentGateway : KlarnaPaymentGatewayBase
    {
        /// <summary>
        /// Processes transaction.
        /// </summary>
        /// <param name="klarnaClient">The klarna client api</param>
        /// <param name="payment">The klarna payment</param>
        /// <param name="message">The message</param>
        /// <returns>True if authorize successful, otherwise false</returns>
        protected override bool ProcessTransaction(Api klarnaClient, KlarnaPayment payment, ref string message)
        {
            // Get the billing address
            var billingAddress = payment.Parent.Parent.OrderAddresses.Cast<OrderAddress>().FirstOrDefault(address2 => address2.Name == payment.BillingAddressId);
            if (billingAddress == null)
            {
                throw new PaymentException(PaymentException.ErrorType.ConfigurationError, "", "Billing address was not specified.");
            }

            // Add line items to pay by Klarna
            KlarnaHelper.AddOrderForm(klarnaClient, payment.Parent);

            // Get the cart
            var cart = payment.Parent.Parent as Cart;
            if (cart == null)
            {
                return false;
            }

            var orderNumber = KlarnaHelper.GenerateOrderNumber(cart.OrderGroupId);
            cart.OrderNumberMethod = new Cart.CreateOrderNumber((c) => orderNumber);

            // Get the SocialSecurityNo from payment
            var socialSecurityNo = payment.SocialSecurityNo as String;
            if (string.IsNullOrEmpty(socialSecurityNo))
            {
                return false;
            }

            // Get the pclassId from payment
            if (string.IsNullOrEmpty(payment.PclassId))
            {
                return false;
            }
            var pclass = int.Parse(payment.PclassId.ToString());
            var status = true;

            string houseNumber = string.Empty;
            string houseExtension = string.Empty;
            try
            {
                switch (billingAddress.CountryCode)
                {
                    case KlarnaHelper.Deu:
                        houseNumber = payment.GermanyHouseNumberText;
                        // House number is required
                        if (string.IsNullOrEmpty(houseNumber))
                        {
                            status = false;
                            break;
                        }
                        break;
                    case KlarnaHelper.Nld:
                        houseNumber = payment.NetherlandsHouseNumberText;
                        houseExtension = payment.NetherlandsHouseExtensionText;
                        // House number is required
                        if (string.IsNullOrEmpty(houseNumber))
                        {
                            status = false;
                            break;
                        }
                        if (houseExtension == null)
                        {
                            houseExtension = string.Empty;
                        }

                        break;
                }

                // Handles gift card and total amount of other active payment methods
                decimal otherActivePaymentAmount = payment.Parent.Total - payment.Amount;
                if (otherActivePaymentAmount != 0)
                {
                    // Klarna doesn't support Gift Card directly so we add special item with negative amount to make the actual amount correct
                    klarnaClient.AddArticle(1, string.Empty, "Gift Card", (double)-otherActivePaymentAmount, 0, 0, GoodsFlags.Handling | GoodsFlags.IncVAT);
                }

                KlarnaHelper.SetKlarnaAddress(klarnaClient, billingAddress, billingAddress, houseNumber, houseExtension);
                klarnaClient.OrderId1 = orderNumber;

                var encoding = KlarnaHelper.GetKlarnaEncodingByCountry(billingAddress.CountryCode);
                InvoiceResponse result = klarnaClient.AddTransaction(socialSecurityNo, null, TransactionFlags.ReturnOcr, pclass, encoding, false);
                // save transaction id from Klarna server for later use
                payment.TransactionID = payment.ProviderTransactionID = result.InvoiceNumber;
            }
            catch (KlarnaException e)
            {
                _logger.Error("Error when processing Klarna part payment request", e);
                ErrorManager.GenerateError(string.Format("Error code: {0}, message: {1}", e.ErrorCode, e.Message));
                status = false;
            }

            return status;
        }

        protected override Api InitKlarnaClient()
        {
            return KlarnaHelper.ConfigKlarna(KlarnaHelper.KlarnaPayment.PartPayment);
        }
    }
}
