﻿using EPiServer.Business.Commerce.Payment.PayPal.Orders;
using EPiServer.Core;
using EPiServer.Framework.Localization;
using EPiServer.Security;
using EPiServer.ServiceLocation;
using EPiServer.Logging;
using Mediachase.Commerce;
using Mediachase.Commerce.Core;
using Mediachase.Commerce.Customers;
using Mediachase.Commerce.Customers.Profile;
using Mediachase.Commerce.Orders;
using Mediachase.Commerce.Orders.Dto;
using Mediachase.Commerce.Orders.Managers;
using Mediachase.Commerce.Plugins.Payment;
using Mediachase.Commerce.Security;
using Mediachase.Commerce.Website.Helpers;
using Mediachase.Data.Provider;
using PayPal.PayPalAPIInterfaceService.Model;
using System;
using System.Collections.Generic;
using System.Data;
using System.IO;
using System.Security.Cryptography;
using System.Text;
using System.Web;
using System.Web.Configuration;
using System.Net;


namespace EPiServer.Business.Commerce.Payment.PayPal
{
    public class PayPalPaymentGateway : AbstractPaymentGateway
    {
        [NonSerialized]
        private readonly ILogger _logger = LogManager.GetLogger(typeof(PayPalPaymentGateway));

        public const string BusinessEmailParameter = "PayPalBusinessEmail";
        public const string AllowChangeAddressParameter = "PayPalChangeAddress";
        public const string AllowGuestParameter = "PayPalAllowGuest";
        public const string PaymentActionParameter = "PayPalPaymentAction";
        public const string UserParameter = "PayPalAPIUser";
        public const string PasswordParameter = "PayPalAPIPassword";
        public const string APISisnatureParameter = "PayPalAPISisnature";
        public const string PALParameter = "PayPalPAL";
        public const string SandBoxParameter = "PayPalSandBox";
        public const string ExpChkoutURLParameter = "PayPalExpChkoutURL";
        public const string SkipConfirmPageParameter = "SkipConfirmPage";
        private OrderAddress _address;
        private const string _hashKey = "@&*PrivateHashKey!%<>?";
        private string _notifyUrl = string.Empty;
        private static object _lock = new object();
        private static DataTable _ISOAlpha;
        private static DataTable _USstates;
        public const string SessionLatestOrderIdKey = "LatestOrderId";
        public const string SessionLatestOrderNumberKey = "LatestOrderNumber";
        public const string SessionLatestOrderTotalKey = "LatestOrderTotal";
        private const string PaymentStatusCompleted = "PayPal payment completed";

        /// <summary>
        /// Gets the PayPalHashKey from AppSettings. 
        /// If this property is set, this hashkey will be use to hash the share token (between our site and PayPal.com, use when call API to PayPal.com)
        /// </summary>
        /// <value>The hash key.</value>
        private static string HashKey
        {
            get
            {
                if (!string.IsNullOrEmpty(WebConfigurationManager.AppSettings["PayPalHashKey"]))
                {
                    return WebConfigurationManager.AppSettings["PayPalHashKey"];
                }
                else
                {
                    return _hashKey;
                }
            }
        }

        #region Get Country, State, Currency from table

        public static List<string> PayPalSupportedCurrencies = new List<string>() { "AUD", "BRL", "CAD",
        "CZK", "DKK", "EUR", "HKD", "HUF", "ILS", "JPY", "MYR",
        "MXN", "NOK", "NZD", "PHP", "PLN", "GBP", "SGD",
        "SEK", "CHF", "TWD", "THB", "USD"
        };

        public static List<string> PayPalSupportedCountriesAndRegions = new List<string>() { "AF", "AX","AL","DZ","AS", "AD", "AO" , "AI" ,"AQ" ,
            "AG" ,"AR" ,"AM" ,"AW" ,"AU" ,"AT" ,"AZ" ,"BS" ,"BH" ,"BD" ,"BB" ,"BY" , "BE" , "BZ" ,"BJ","BM","BT","BO","BA" ,
            "BW" ,"BV" ,"BR" ,"IO" ,"BN" ,"BG" ,"BF" ,"BI" ,"KH" ,"CM" ,"CA" ,"CV" ,"KY" ,"CF" ,"TD" ,"CL" ,"CN" ,"CX" ,"CC" ,"CO" ,"KM" ,"CG" ,"CD" ,
            "CK" ,"CR" ,"CI" ,"HR" ,"CU" ,"CY", "CZ" ,"DK", "DJ" ,"DM" ,"DO" ,"EC" ,"EG" ,"SV" ,"GQ" ,"ER" ,"EE" ,"ET" ,"FK" ,"FO" ,"FJ" ,"FI" ,"FR" ,
            "GF" ,"PF" ,"TF" ,"GA" ,"GM" ,"GE" ,"DE" ,"GH" ,"GI" ,"GR" ,"GL" ,"GD" ,"GP" ,"GU" ,"GT" ,"GG" ,"GN" ,"GW" ,"GY" ,"HT" ,"HM" ,"VA" ,"HN" ,
            "HK" ,"HU" ,"IS" ,"IN" ,"ID" ,"IR" ,"IQ" ,"IE" ,"IM" ,"IL" ,"IT" ,"JM" ,"JP" ,"JE" ,"JO" ,"KZ" ,"KE" ,"KI" ,"KP" ,"KR" ,"KW" ,"KG", "LA" ,
            "LV" ,"LB" ,"LS" ,"LR" ,"LY" ,"LI" ,"LT" ,"LU" ,"MO" ,"MK" ,"MG" ,"MW" ,"MY" ,"MV" ,"ML" ,"MT" ,"MH" ,"MQ" ,"MR" ,"MU" ,"YT" ,"MX","FM" ,"MD" ,
            "MC" ,"MN" ,"MS" ,"MA" ,"MZ" ,"MM" ,"NA" ,"NR" ,"NP" ,"NL" ,"AN" ,"NC" ,"NZ" ,"NI" ,"NE" ,"NG" ,"NU" ,"NF" ,"MP" ,"NO" ,"OM" ,"PK" ,"PW" ,"PS" ,
            "PA" ,"PG" ,"PY" ,"PE" ,"PH" ,"PN" ,"PL" ,"PT" ,"PR" ,"QA" ,"RE" ,"RO", "RU" ,"RW" ,"SH" ,"KN" ,"LC" ,"PM" ,"VC" ,"WS" ,"SM" ,"ST" ,"SA" ,"SN" ,
            "CS" ,"SL","SG" ,"SK" ,"SI" ,"SB","SO" ,"ZA" ,"GS" ,"ES" ,"LK" ,"SD","SR","SJ","SZ" ,"SE" ,"CH" ,"SY","TW" ,"TJ" ,"TZ" ,"TH" ,"TL" ,"TG" ,"TK" ,
            "TO" ,"TT" ,"TN" ,"TR" ,"TM" ,"TC" ,"TV" ,"UG" ,"UA" ,"AE" ,"GB" ,"US" ,"UM" ,"UY" ,"UZ" ,"VU" ,"VE" ,"VN" ,"VG" ,"VI" ,"WF" ,"EH" ,"YE" ,"ZM" , "ZW" };

        // TODO change to the appropriate path of the template.
        internal static readonly string IsoFilePath = HttpContext.Current.Server.MapPath("~/Templates/Sample/Units/CartCheckout/PayPal/");

        /// <summary>
        /// Gets the country code table, to changes between ISO Alpha2 and Alpha3 contry code
        /// </summary>
        /// <returns></returns>
        /// <value>The country code table.</value>
        public DataTable CountryCodeTable
        {
            get
            {
                if (_ISOAlpha == null)
                {
                    lock (_lock)
                    {
                        if (_ISOAlpha == null)
                        {
                            string isoFilePath = IsoFilePath;
                            if (isoFilePath.EndsWith(@"\"))
                            {
                                isoFilePath += "ISOCodes.txt";
                            }
                            else
                            {
                                isoFilePath += @"\ISOCodes.txt";
                            }

                            try
                            {
                                using (StreamReader reader = new StreamReader(isoFilePath))
                                {
                                    _ISOAlpha = new DataTable();
                                    _ISOAlpha.Columns.Add("ISO3", typeof(string));
                                    _ISOAlpha.Columns.Add("ISO2", typeof(string));
                                    _ISOAlpha.Columns.Add("Name", typeof(string));

                                    string s = reader.ReadLine();
                                    while (!reader.EndOfStream && !string.IsNullOrEmpty(s))
                                    {
                                        string[] values = s.Split(new char[] { ' ' }, StringSplitOptions.RemoveEmptyEntries);
                                        string alpha3 = values[0];
                                        string alpha2 = values[1];
                                        string name = string.Empty;
                                        for (int i = 2; i < values.Length; i++)
                                        {
                                            name += " " + values[i];
                                        }
                                        _ISOAlpha.Rows.Add(new object[] { alpha3, alpha2, name.TrimStart(' ') });
                                        s = reader.ReadLine();
                                    }
                                }
                            }
                            catch
                            {
                                // cannot read from iso mapping file, file does not exist or content is not wellform
                            }
                        }
                    }
                }
                return _ISOAlpha;
            }
        }

        /// <summary>
        /// Gets the US and Canadian state codes, for
        /// US and Canadian, PayPal accepts only two letter state code.
        /// </summary>
        /// <value>The U sstates.</value>
        public DataTable USstates
        {
            get
            {
                if (_USstates == null)
                {
                    lock (_lock)
                    {
                        if (_USstates == null)
                        {
                            _USstates = new DataTable();
                            _USstates.Columns.Add("Name", typeof(string));
                            _USstates.Columns.Add("Code", typeof(string));

                            string isoFilePath = IsoFilePath;

                            if (isoFilePath.EndsWith(@"\"))
                            {
                                isoFilePath += "CanadianOrUSstates.txt";
                            }
                            else
                            {
                                isoFilePath += @"\CanadianOrUSstates.txt";
                            }
                            try
                            {
                                using (StreamReader reader = new StreamReader(isoFilePath))
                                {
                                    while (!reader.EndOfStream)
                                    {
                                        string name = reader.ReadLine();
                                        string code = reader.ReadLine();
                                        _USstates.Rows.Add(new object[] { name, code });
                                    }
                                }
                            }
                            catch
                            {

                            }
                        }
                    }
                }
                return _USstates;
            }
        }

        /// <summary>
        /// Gets the alpha3 country code.
        /// </summary>
        /// <param name="countryCode">The country code.</param>
        /// <returns></returns>
        public string GetAlpha3CountryCode(string countryCode)
        {
            string name = string.Empty;
            foreach (DataRow row in CountryCodeTable.Rows)
            {
                if (row[1].ToString().ToUpperInvariant() == countryCode.ToUpperInvariant())
                {
                    name = row[0].ToString().ToUpperInvariant();
                    break;
                }
            }
            return name;
        }

        /// <summary>
        /// Gets the alpha2 country code.
        /// </summary>
        /// <param name="address">The address.</param>
        /// <returns></returns>
        public CountryCodeType GetAlpha2CountryCode(OrderAddress address)
        {
            if (CountryCodeTable == null)
            {
                return CountryCodeType.CUSTOMCODE;
            }

            string code = string.Empty;
            foreach (DataRow row in CountryCodeTable.Rows)
            {
                if (row[0].ToString().Equals(address.CountryCode, StringComparison.OrdinalIgnoreCase))
                {
                    code = row[1].ToString().ToUpperInvariant();
                    break;
                }
            }

            foreach (CountryCodeType value in Enum.GetValues(typeof(CountryCodeType)))
            {
                if (value.ToString().ToUpperInvariant() == code)
                {
                    return value;
                }
            }

            return CountryCodeType.CUSTOMCODE;
        }

        /// <summary>
        /// States the name from state code.
        /// </summary>
        /// <param name="stateCode">The state code.</param>
        /// <returns></returns>
        public string GetStateName(string stateCode)
        {
            if (string.IsNullOrEmpty(stateCode))
            {
                return stateCode;
            }
            foreach (DataRow row in USstates.Rows)
            {
                if (string.Equals(row[1].ToString(), stateCode, StringComparison.OrdinalIgnoreCase))
                {
                    return row[0].ToString();
                }
            }
            return stateCode;
        }

        /// <summary>
        /// States the code from the state name
        /// </summary>
        /// <param name="name">The name.</param>
        /// <returns></returns>
        public string GetStateCode(string name)
        {
            if (string.IsNullOrEmpty(name))
            {
                return name;
            }
            foreach (DataRow row in USstates.Rows)
            {
                if (string.Equals(row[0].ToString(), name, StringComparison.OrdinalIgnoreCase))
                {
                    return row[1].ToString().ToUpperInvariant();
                }
            }
            return name;
        }

        /// <summary>
        /// Try to get the PayPal's currency code from payment's parent BillingCurrency or SiteContext's currencyCode
        /// </summary>
        /// <value>The PayPal currency code.</value>
        public CurrencyCodeType GetCurrencyCode(Mediachase.Commerce.Orders.Payment payment)
        {
            CurrencyCodeType currency = CurrencyCodeType.CUSTOMCODE;
            if (string.IsNullOrEmpty(payment.Parent.Parent.BillingCurrency))
            {
                currency = GetCurrencyCode(SiteContext.Current.Currency);
            }
            else
            {
                currency = GetCurrencyCode(payment.Parent.Parent.BillingCurrency);
            }
            return currency;
        }

        /// <summary>
        /// Currency code for PayPal
        /// </summary>
        /// <param name="currency">The currency.</param>
        /// <returns></returns>
        private CurrencyCodeType GetCurrencyCode(Currency currency)
        {
            string cur = currency.CurrencyCode.ToUpperInvariant();
            if (PayPalPaymentGateway.PayPalSupportedCurrencies.Contains(cur))
            {
                foreach (CurrencyCodeType value in Enum.GetValues(typeof(CurrencyCodeType)))
                {
                    if (value.ToString().ToUpperInvariant() == cur)
                    {
                        return value;
                    }
                }
            }
            return CurrencyCodeType.CUSTOMCODE;
        }
        #endregion





        /// <summary>
        /// Main entry point of ECF Payment Gateway.
        /// </summary>
        /// <param name="payment">The payment to process</param>
        /// <param name="message">The message.</param>
        /// <returns>return false and set the message will make the WorkFlow activity raise PaymentExcetion(message)</returns>
        public override bool ProcessPayment(Mediachase.Commerce.Orders.Payment payment, ref string message)
        {
            #region pre-condition checking
            
            if (HttpContext.Current == null)
            {
                message = Utilities.Translate("ProcessPaymentNullHttpContext");
                return true;
            }

            #endregion

            ServicePointManager.SecurityProtocol = SecurityProtocolType.Tls12;

            Cart cart = payment.Parent.Parent as Cart;
            // the order which is created by Commerce Manager
            if (cart == null && payment.Parent.Parent is PurchaseOrder)
            {
                #region CAPTURE

                if (payment.TransactionType == TransactionType.Capture.ToString())
                {
                    return ProcessPaymentCapture(payment, ref message);
                }

                #endregion

                #region REFUND

                // When "Refund" shipment in Commerce Manager, this method will be invoked with the TransactionType is Credit
                if (payment.TransactionType == TransactionType.Credit.ToString())
                {
                    return ProcessPaymentRefund(payment, ref message);
                }

                #endregion

                #region CREATE ORDER IN COMMERCE MANAGER
                
                // right now we do not support processing the order which is created by Commerce Manager
                message = "The current payment method does not support this order type.";
                return false;   // raise exception

                #endregion
            }

            #region PAID ON PAYPAL
            
            if (cart != null && cart.Status == PaymentStatusCompleted)
            {
                // return true because this shopping cart has been paid already on PayPal
                // when program flow redirects back from PayPal to PayPal.aspx file, call ProcessSuccessfulTransaction, run WorkFlow
                message = Utilities.Translate("ProcessPaymentStatusCompleted");
                return true;
            }

            #endregion

            
            // CHECKOUT
            return ProcessPaymentCheckout(payment, ref message, cart);
        }

        /// <summary>
        /// Captures an authorized payment.
        /// </summary>
        /// <remarks>
        /// <para>
        /// See API doc here https://developer.paypal.com/webapps/developer/docs/classic/api/merchant/DoCapture_API_Operation_SOAP/
        /// </para>
        /// <param name="payment">The payment to process</param>
        /// <param name="message">The message.</param>
        /// <returns>return false and set the message will make the WorkFlow activity raise PaymentExcetion(message)</returns>
        private bool ProcessPaymentCapture(Mediachase.Commerce.Orders.Payment payment, ref string message)
        {
            // Implement refund feature logic for current payment gateway
            var captureAmount = payment.Amount;
            var po = (payment.Parent.Parent as PurchaseOrder);
            if (po == null || captureAmount <= 0)
            {
                message = "Nothing to capture";
                return false;
            }
            var currency = GetCurrencyCode(payment);

            var captureRequest = new DoCaptureRequestType();
            captureRequest.AuthorizationID = payment.TransactionID;  // original transactionID (which PayPal gave us when DoExpressCheckoutPayment, DoDirectPayment, or CheckOut)
            captureRequest.Amount = captureAmount.ToPayPalAmount(currency);   // if refund with Partial, we have to set the Amount
            
            if (payment.Amount >= po.Total)
            {
                captureRequest.CompleteType = CompleteCodeType.COMPLETE;
            }
            else
            {
                captureRequest.CompleteType = CompleteCodeType.NOTCOMPLETE;
            }
            captureRequest.InvoiceID = po.TrackingNumber;

            captureRequest.Note = string.Format("[{2}-{3}] captured {0}{1} for [PurchaseOrder-{4}]",
                captureAmount, currency,
                payment.PaymentMethodName, payment.TransactionID,
                po.TrackingNumber
                );


            var caller = Utilities.GetPayPalAPICallerServices();
            var doCaptureReq = new DoCaptureReq();
            doCaptureReq.DoCaptureRequest = captureRequest;
            var captureResponse = (DoCaptureResponseType)caller.DoCapture(doCaptureReq);

            string errorCheck = captureResponse.CheckErrors();
            if (!string.IsNullOrEmpty(errorCheck))
            {
                message = errorCheck;
                return false;
            }
            else
            {
                var captureResponseDetails = captureResponse.DoCaptureResponseDetails;
                var paymentInfo = captureResponseDetails.PaymentInfo;
                
                // Extract the response details.
                payment.TransactionID = paymentInfo.TransactionID;
                payment.Status = paymentInfo.PaymentStatus.ToString();
                payment.AcceptChanges();

                message = string.Format("[{0}] [Capture payment-{1}] [Status: {6}] .Response: {2} at Timestamp={3}: {4}{5}",
                    payment.PaymentMethodName,
                    paymentInfo.TransactionID,
                    captureResponse.Ack.ToString(), captureResponse.Timestamp.ToString(),
                    paymentInfo.GrossAmount.value, paymentInfo.GrossAmount.currencyID,
                    paymentInfo.PaymentStatus.ToString()
                );

                // add a new order note about this refund
                var orderNote = po.OrderNotes.AddNew();
                orderNote.Created = DateTime.Now;
                orderNote.CustomerId = po.CustomerId;
                orderNote.Detail = message;
                orderNote.Title = "CAPTURE";
                orderNote.Type = OrderNoteTypes.System.ToString();
                po.AcceptChanges();

                return true;
            }
        }


        /// <summary>
        /// Process payment when a refund request happens
        /// </summary>
        /// <remarks>
        /// <para>
        /// See API doc here https://www.x.com/developers/paypal/documentation-tools/api/refundtransaction-api-operation-soap
        /// </para>
        /// <para>
        /// You may offer a refund only for a limited time, usually 60 days. If you need to make a refund after that time, you will need to initiate a new PayPal payment to your buyer.
        /// If you offer the buyer a partial refund, she has 10 days to decline it if she wishes. (Full refunds are automatically processed.)
        /// </para>
        /// </remarks>
        /// <param name="payment">The payment to process</param>
        /// <param name="message">The message.</param>
        /// <returns>return false and set the message will make the WorkFlow activity raise PaymentExcetion(message)</returns>
        private bool ProcessPaymentRefund(Mediachase.Commerce.Orders.Payment payment, ref string message)
        {
            // Implement refund feature logic for current payment gateway
            var refundAmount = payment.Amount;
            var po = (payment.Parent.Parent as PurchaseOrder);
            if (po == null || refundAmount <= 0)
            {
                message = "Nothing to refund";  // TODO: localize
                return false;
            }

            var currency = GetCurrencyCode(payment);
            //payment.Parent.Status;
            var invoiceNumber = po.TrackingNumber;
            var ppTransactionID = payment.TransactionID;

            #region call payment gateway API to do refund business

            // Create the Refund Request
            var refundRequest = new RefundTransactionRequestType();
            refundRequest.TransactionID = ppTransactionID;  // original transactionID (which payPal gave us when do ExpressCheckout)
            refundRequest.Memo = string.Format("[{2}-{3}] refunds {0}{1} for [PurchaseOrder-{4}]", 
                refundAmount, po.BillingCurrency,
                payment.PaymentMethodName, payment.TransactionID, 
                invoiceNumber
                );


            // NOTE: If RefundType is Full, do not set the amount.
            // refundRequest.RefundType = RefundType.Full;  //refund a full or partial amount
            refundRequest.RefundType = RefundType.PARTIAL;  //refund a partial amount
            refundRequest.Amount = refundAmount.ToPayPalAmount(currency);   // if refund with Partial, we have to set the Amount

            var caller = Utilities.GetPayPalAPICallerServices();
            var refundResponse = (RefundTransactionResponseType)caller.RefundTransaction(new RefundTransactionReq() { RefundTransactionRequest = refundRequest });
            string errorCheck = refundResponse.CheckErrors();
            if (!string.IsNullOrEmpty(errorCheck))
            {
                message = errorCheck;
                return false;
            }
            else
            {
                // Extract the response details.
		        payment.TransactionID = refundResponse.RefundTransactionID;
                payment.AcceptChanges();

                message = string.Format("[{0}] [RefundTransaction-{1}] Response: {2} at Timestamp={3}: {4}{5}",
                    payment.PaymentMethodName,
                    refundResponse.RefundTransactionID,
                    refundResponse.Ack.ToString(), refundResponse.Timestamp.ToString(),
                    refundResponse.GrossRefundAmount.value, refundResponse.GrossRefundAmount.currencyID
                );

                // add a new order note about this refund
                var orderNote = po.OrderNotes.AddNew();
                orderNote.Created = DateTime.Now;
                orderNote.CustomerId = po.CustomerId;
                orderNote.Detail = message;
                orderNote.Title = "REFUND"; 
                orderNote.Type = OrderNoteTypes.System.ToString();
                po.AcceptChanges();
                
                return true;
            }

            #endregion

        }


        /// <summary>
        /// Process payment when user checkout
        /// </summary>
        /// <param name="payment">The payment to process</param>
        /// <param name="message">The message.</param>
        /// <param name="cart">The current cart.</param>
        /// <returns>return false and set the message will make the WorkFlow activity raise PaymentExcetion(message)</returns>
        private bool ProcessPaymentCheckout(Mediachase.Commerce.Orders.Payment payment, ref string message, Cart cart)
        {
            var payPalPayment = payment as PayPalPayment;
            string orderNumberID = GenerateOrderNumber(cart.OrderGroupId);

            #region Build notify Url (PayPal will redirect back to this url after processing OK or FAIL)

            // This key is sent to PayPal using https so it is not likely it will come from other because 
            // only PayPal knows this key to send back to us
            string acceptSecurityKey = GetMD5Key(orderNumberID + "accepted");
            string cancelSecurityKey = GetMD5Key(orderNumberID + "canceled");

            _notifyUrl = UriSupport.AbsoluteUrlBySettings(Utilities.GetUrlValueFromStartPage("PayPalPaymentPage"));

            string acceptUrl = UriSupport.AddQueryString(_notifyUrl, "accept", "true");
            acceptUrl = UriSupport.AddQueryString(acceptUrl, "hash", acceptSecurityKey);

            var cancelUrl = HttpContext.Current != null && HttpContext.Current.Request.UrlReferrer != null ? 
                                    HttpContext.Current.Request.UrlReferrer.ToString() : 
                                    DataFactory.Instance.GetPage(PageReference.StartPage).LinkURL;
            cancelUrl = UriSupport.AddQueryString(cancelUrl, "accept", "false");
            cancelUrl = UriSupport.AddQueryString(cancelUrl, "hash", cancelSecurityKey);

            #endregion

            #region Create CallerServices and fill its APIProfile with Commerce's PayPal security credentials

            var caller = Utilities.GetPayPalAPICallerServices();
            if (caller == null || string.IsNullOrEmpty(this.Settings[PayPalPaymentGateway.ExpChkoutURLParameter]))
            {
                Utilities.UpdateCartInstanceId(cart);
                message = Utilities.Translate("PayPalSettingsError");
                return false;   // raise exception
            }

            #endregion

            #region Create the PayPal API SOAP request object.

            SetExpressCheckoutRequestType setChkOutrequest = new SetExpressCheckoutRequestType();

            // Add request-specific fields to the request.
            // Create the request details object.
            setChkOutrequest.SetExpressCheckoutRequestDetails = new SetExpressCheckoutRequestDetailsType();
            if (this.Settings[PaymentActionParameter] == "Authorization")
            {
                setChkOutrequest.SetExpressCheckoutRequestDetails.PaymentAction = PaymentActionCodeType.AUTHORIZATION;
            }
            else if (this.Settings[PaymentActionParameter] == "Sale")
            {
                setChkOutrequest.SetExpressCheckoutRequestDetails.PaymentAction = PaymentActionCodeType.SALE;
            }

            if (this.Settings[AllowChangeAddressParameter] != "1")
            {
                setChkOutrequest.SetExpressCheckoutRequestDetails.AddressOverride = "1";
            }
            setChkOutrequest.SetExpressCheckoutRequestDetails.BillingAddress = GetBillingAddress(payPalPayment);
            setChkOutrequest.SetExpressCheckoutRequestDetails.BuyerEmail = _address.Email;
            setChkOutrequest.SetExpressCheckoutRequestDetails.PaymentDetails = new List<PaymentDetailsType>() { GetPaymentDetailsType(payPalPayment, orderNumberID) };
            if (this.Settings[AllowGuestParameter] == "1")
            {
                setChkOutrequest.SetExpressCheckoutRequestDetails.SolutionType = SolutionTypeType.SOLE;
                setChkOutrequest.SetExpressCheckoutRequestDetails.LandingPage = LandingPageType.BILLING;
            }
            setChkOutrequest.SetExpressCheckoutRequestDetails.CancelURL = cancelUrl;
            setChkOutrequest.SetExpressCheckoutRequestDetails.ReturnURL = acceptUrl;

            #endregion

            #region Execute the API operation and obtain the response.

            SetExpressCheckoutResponseType setChkOutResponse = (SetExpressCheckoutResponseType)caller.SetExpressCheckout(new SetExpressCheckoutReq() { SetExpressCheckoutRequest = setChkOutrequest });
            string errorCheck = setChkOutResponse.CheckErrors();
            if (!string.IsNullOrEmpty(errorCheck))
            {
                // TODO: Utilities.UpdateCartInstanceId(cart);
                message = !string.IsNullOrEmpty(errorCheck) ? errorCheck : Utilities.Translate("ProcessPaymentSetExpressCheckoutError");
                _logger.Error(message);
                return false;   // raise exception
            }
            
            // validation checking with PayPal OK (Server's PayPal API, Billing Address, Shipping Address, ... do redirect to PayPal.com
            string redirectUrl = this.Settings[PayPalPaymentGateway.ExpChkoutURLParameter];
            redirectUrl = UriSupport.AddQueryString(redirectUrl, "cmd", "_express-checkout");
            redirectUrl = UriSupport.AddQueryString(redirectUrl, "token", setChkOutResponse.Token);
            redirectUrl = UriSupport.AddQueryString(redirectUrl, "useraction", "commit");

            payPalPayment.PayPalOrderNumber = orderNumberID;
            payPalPayment.PayPalExpToken = setChkOutResponse.Token;

            Utilities.UpdateCartInstanceId(cart);
            message = string.Format("---PayPal-SetExpressCheckout is successful. Redirect end user to {0}", redirectUrl);
            _logger.Information(message);
            RedirectParentPage(redirectUrl);

            #endregion

            return true;
        }



        /// <summary>
        /// Processes the successful transaction, call by NotifyPage (PayPal.aspx) when PayPal.com redirect back.
        /// </summary>
        /// <param name="acceptUrl">redirect to this url when everything is OK</param>
        /// <param name="cancelUrl">redirect to this url when error happens</param>
        /// <param name="payPalPayment">PayPal payment object, which contains Order number and Exp Token.</param>
        public void ProcessSuccessfulTransaction(string acceptUrl, string cancelUrl, PayPalPayment payPalPayment)
        {
            #region pre-condition checking

            if (HttpContext.Current == null)
            {
                return;
            }

            if (HttpContext.Current.Session != null)
            {
                HttpContext.Current.Session.Remove("LastCouponCode");
            }

            CartHelper helper = new CartHelper(Cart.DefaultName);
            Cart cart = helper.Cart;
            if (cart == null)
            {
                //return to the shopping cart page immediately and show error messages
                ProcessUnsuccessfulTransaction(cancelUrl, Utilities.Translate("CommitTranErrorCartNull"));
            }

            #endregion

            var email = helper.PrimaryAddress == null ? string.Empty : helper.PrimaryAddress.Email;

            // Make sure to execute within transaction
            using (TransactionScope scope = new TransactionScope())
            {
                PaymentMethodDto payPal = Utilities.GetPayPalPaymentMethod();
                
                var caller = Utilities.GetPayPalAPICallerServices();

                #region Create the GetExpressCheckoutDetails PayPal API request object.

                GetExpressCheckoutDetailsRequestType getDetailRequest = new GetExpressCheckoutDetailsRequestType();

                // Add request-specific fields to the request.
                getDetailRequest.Token = payPalPayment.PayPalExpToken;

                // Execute the API operation and obtain the response.
                var getDetailsResponse = new GetExpressCheckoutDetailsResponseType();
                getDetailsResponse = (GetExpressCheckoutDetailsResponseType)caller.GetExpressCheckoutDetails(new GetExpressCheckoutDetailsReq() { GetExpressCheckoutDetailsRequest = getDetailRequest });
                string errorCheck = getDetailsResponse.CheckErrors();
                if (!string.IsNullOrEmpty(errorCheck))
                {
                    // TODO: Utilities.UpdateCartInstanceId(cart);
                    // unsuccessful getdetail call
                    ProcessUnsuccessfulTransaction(cancelUrl, !string.IsNullOrEmpty(errorCheck) ? errorCheck : Utilities.Translate("GetExpressCheckoutDetailsError"));
                }

                // get commerceOrderId from what we put to PayPal instead of getting from cookie
                payPalPayment.PayPalOrderNumber = getDetailsResponse.GetExpressCheckoutDetailsResponseDetails.InvoiceID;

                //process details sent from paypal, changing addresses if required
                string billAddEmptyMsg = null, shipAddEmptyMsg = null;
                AddressType newShippingAddressFromPayPal = null;
                OrderAddress modifiedBillingAddress = null, modifiedShippingAddress = null;
                ProcessExChkOutDetails(getDetailsResponse, cart, ref billAddEmptyMsg, ref shipAddEmptyMsg, ref modifiedBillingAddress,
                    ref modifiedShippingAddress, ref newShippingAddressFromPayPal);

                #endregion


                #region Create the DoExpressCheckoutPayment PayPal API request object, to complete the transaction
                
                // Create the request object.
                var doCheckOutRequest = new DoExpressCheckoutPaymentRequestType();

                // Add request-specific fields to the request.
                // Create the request details object.
                doCheckOutRequest.DoExpressCheckoutPaymentRequestDetails = new DoExpressCheckoutPaymentRequestDetailsType();
                doCheckOutRequest.DoExpressCheckoutPaymentRequestDetails.Token = payPalPayment.PayPalExpToken;
                doCheckOutRequest.DoExpressCheckoutPaymentRequestDetails.PayerID = getDetailsResponse.GetExpressCheckoutDetailsResponseDetails.PayerInfo.PayerID;
                if (payPal.GetParameterValueByName(PaymentActionParameter) == "Authorization")
                {
                    doCheckOutRequest.DoExpressCheckoutPaymentRequestDetails.PaymentAction = PaymentActionCodeType.AUTHORIZATION;
                }
                else if (payPal.GetParameterValueByName(PaymentActionParameter) == "Sale")
                {
                    doCheckOutRequest.DoExpressCheckoutPaymentRequestDetails.PaymentAction = PaymentActionCodeType.SALE;
                }

                PaymentDetailsType sentDetails = GetPaymentDetailsType(cart.OrderForms[0].Payments[0], payPalPayment.PayPalOrderNumber);
                doCheckOutRequest.DoExpressCheckoutPaymentRequestDetails.PaymentDetails = new List<PaymentDetailsType>() { sentDetails };
                if (newShippingAddressFromPayPal != null)
                {
                    doCheckOutRequest.DoExpressCheckoutPaymentRequestDetails.PaymentDetails[0].ShipToAddress = newShippingAddressFromPayPal;
                }
                // Execute the API operation and obtain the response.
                DoExpressCheckoutPaymentResponseType doCheckOutResponse = new DoExpressCheckoutPaymentResponseType();
                doCheckOutResponse = (DoExpressCheckoutPaymentResponseType)caller.DoExpressCheckoutPayment(new DoExpressCheckoutPaymentReq() { DoExpressCheckoutPaymentRequest = doCheckOutRequest });

                errorCheck = doCheckOutResponse.CheckErrors();
                if (!string.IsNullOrEmpty(errorCheck))
                {
                    // TODO: Utilities.UpdateCartInstanceId(cart);
                    // unsuccessful doCheckout response
                    ProcessUnsuccessfulTransaction(cancelUrl, !string.IsNullOrEmpty(errorCheck) ? errorCheck : Utilities.Translate("CommitTranErrorDoExpressCheckoutPayment"));
                }

                #endregion

                // everything is fine, this is a flag to tell ProcessPayment know about this case: redirect back from PayPal with accepted payment
                cart.Status = PaymentStatusCompleted;

                // Change status of payments to processed. 
                // It must be done before execute workflow to ensure payments which should mark as processed.
                // To avoid get errors when executed workflow.
                foreach (OrderForm orderForm in cart.OrderForms)
                {
                    foreach (Mediachase.Commerce.Orders.Payment payment in orderForm.Payments)
                    {
                        if (payment != null)
                        {
                            PaymentStatusManager.ProcessPayment(payment);
                        }
                    }
                }

                // Execute CheckOutWorkflow with parameter to ignore running process payment activity again.
                var isIgnoreProcessPayment = new Dictionary<string, object>();
                isIgnoreProcessPayment.Add("PreventProcessPayment", true);
                OrderGroupWorkflowManager.RunWorkflow(cart, OrderGroupWorkflowManager.CartCheckOutWorkflowName, true, isIgnoreProcessPayment);

                // Saving addresses for logined users
                TrySaveAddressForLoginedUser(modifiedShippingAddress, modifiedBillingAddress, cart);

                // Save changes
                cart.OrderNumberMethod = new Cart.CreateOrderNumber((c) => payPalPayment.PayPalOrderNumber);
                PurchaseOrder po = cart.SaveAsPurchaseOrder();  // NOTE: this might cause problem when checkout using multiple shipping address (because ECF workflow does not handle it. Modify the workflow instead of modify in this payment)
                // NOTE: po.TrackingNumber hold the orderNumber

                // read PayPal's transactionId here, in order to store it, refund ...
                var ppTransactionID = doCheckOutResponse.DoExpressCheckoutPaymentResponseDetails.PaymentInfo[0].TransactionID;
                UpdateTransactionIdOfPaymentMethod(po, payPal, ppTransactionID);

                // Update last order datetime for CurrentContact
                UpdateLastOrderOfCurrentContact(CustomerContext.Current.CurrentContact, po.Created);

                AddNoteToPurchaseOrder("New order placed by {0} in {1}", po, PrincipalInfo.CurrentPrincipal.Identity.Name, "Public site");                
                
                // Remove old cart
                cart.Delete();
                cart.AcceptChanges();

                // Update display name of product by current language
                po.UpdateDisplayNameWithCurrentLanguage();

                po.AcceptChanges();

                // Commit changes
                scope.Complete();
                acceptUrl = UriSupport.AddQueryString(acceptUrl, "success", "true");
                acceptUrl = UriSupport.AddQueryString(acceptUrl, "ordernumber", payPalPayment.PayPalOrderNumber);
                acceptUrl = UriSupport.AddQueryString(acceptUrl, "email", email);
                
            }   // end ECF transaction scope


            _logger.Information(String.Format("PayPal transaction succeeds, redirect end user to {0}", acceptUrl));
            HttpContext.Current.Response.Redirect(acceptUrl, true);
        }

        /// <summary>
        /// Adds the note to purchase order.
        /// </summary>
        /// <param name="note">Name of the note.</param>
        /// <param name="purchaseOrder">The purchase order.</param>
        /// <param name="parmeters">The parmeters.</param>
        protected void AddNoteToPurchaseOrder(string note, PurchaseOrder purchaseOrder, params string[] parmeters)
        {
            var noteDetail = String.Format(note, parmeters);
            OrderNotesManager.AddNoteToPurchaseOrder(purchaseOrder, noteDetail, OrderNoteTypes.System, PrincipalInfo.CurrentPrincipal.GetContactId());
        }

        /// <summary>
        /// loop through all payments in the PurchaseOrder, find payment of paymentMethod type, set TransactionId
        /// </summary>
        /// <param name="po">PurchaseOrder need to update</param>
        /// <param name="paymentMethod">e.g. PayPal</param>
        /// <param name="paymentGatewayTransactionID">transactionId from PaymentGateway, e.g. : from PayPal TransactionID</param>
        private void UpdateTransactionIdOfPaymentMethod(PurchaseOrder po, PaymentMethodDto paymentMethod, string paymentGatewayTransactionID)
        {
            if (paymentMethod == null || paymentMethod.PaymentMethod.Rows.Count <= 0)
            {
                return;
            }

            // take from first row
            var guidPaymentMethodId = (Guid)paymentMethod.PaymentMethod.Rows[0][paymentMethod.PaymentMethod.PaymentMethodIdColumn];

            // loop through all payments in the PurchaseOrder, find payment with id equal guidPaymentMethodId, set TransactionId
            foreach (Mediachase.Commerce.Orders.OrderForm orderForm in po.OrderForms)
            {
                foreach (Mediachase.Commerce.Orders.Payment payment in orderForm.Payments)
                {
                    if (payment.PaymentMethodId.Equals(guidPaymentMethodId))
                    {
                        payment.TransactionID = paymentGatewayTransactionID;
                        payment.AcceptChanges();
                    }
                }
            }
        }

        
        /// <summary>
        /// If current user is not anonymous, try to save the ShippingAddress and BillingAddress to its Contact
        /// </summary>
        /// <param name="modifiedShippingAddress">shipping address to save</param>
        /// <param name="modifiedBillingAddress">billing address to save</param>
        /// <param name="cart">if modifiedShippingAddress and/or modifiedBillingAddress is null, try to take addresses from Cart.OrderForm</param>
        private void TrySaveAddressForLoginedUser(OrderAddress modifiedShippingAddress, OrderAddress modifiedBillingAddress, Cart cart)
        {
            if (HttpContext.Current == null)
            {
                return;
            }
            var httpProfile = HttpContext.Current.Profile;
            var profile = httpProfile == null ? null : new CustomerProfileWrapper(httpProfile);
            if (profile != null && !profile.IsAnonymous)
            {
                //the new shipping address from paypal, save it as a new customer address
                OrderAddress shipAdd = modifiedShippingAddress ??
                                       new CartHelper(cart).FindAddressByName(
                                           cart.OrderForms[0].Shipments[0].ShippingAddressId);
                shipAdd.AcceptChanges();
                SaveCustomerAddress(shipAdd, CustomerAddressTypeEnum.Shipping);

                //the new billing address from PayPal, as it as new customer address
                OrderAddress billAdd = modifiedBillingAddress ??
                                       new CartHelper(cart).FindAddressByName(cart.OrderForms[0].BillingAddressId);
                billAdd.AcceptChanges();
                SaveCustomerAddress(billAdd, CustomerAddressTypeEnum.Billing);
            }
        }

        /// <summary>
        /// Update last order timestamp which current user completed
        /// </summary>
        /// <param name="contact"></param>
        /// <param name="datetime"></param>
        private void UpdateLastOrderOfCurrentContact(CustomerContact contact, DateTime datetime)
        {
            if (contact != null)
            {
                contact.LastOrder = datetime;
                contact.SaveChanges();
            }
        }


        /// <summary>
        /// Processes the unsuccessful transaction, delete PayPal-related information in Cookie.
        /// Redirect to the <paramref name="cancelUrl"/>
        /// </summary>
        /// <param name="cancelUrl"></param>
        /// <param name="errorMessage"></param>
        public void ProcessUnsuccessfulTransaction(string cancelUrl, string errorMessage)
        {
            if (HttpContext.Current == null)
            {
                return;
            }

            _logger.Error(String.Format("PayPal transaction failed [{0}], redirect end user to {1}", errorMessage, cancelUrl));
            HttpContext.Current.Response.Redirect(cancelUrl, true);
        }

        /// <summary>
        /// Saves the new order address to a (new) customer contact if not already existed.
        /// </summary>
        /// <param name="address">The address to save</param>
        /// <param name="type">Customer address type</param>
        private void SaveCustomerAddress(OrderAddress address, CustomerAddressTypeEnum type)
        {
            // Add to contact address
            var customerContact = PrincipalInfo.CurrentPrincipal.GetCustomerContact();
            if (customerContact != null)
            {
                CustomerAddress customerAddress = StoreHelper.ConvertToCustomerAddress(address);
                if (customerContact.ContactAddresses == null || !StoreHelper.IsAddressInCollection(customerContact.ContactAddresses, customerAddress))
                {
                    customerAddress.AddressType = type;
                    customerAddress.Name = customerAddress.Name.StripPreviewText(46);
                    customerAddress.FirstName = customerAddress.FirstName.StripPreviewText(60);
                    customerAddress.LastName = customerAddress.LastName.StripPreviewText(60);
                    customerAddress.Line1 = customerAddress.Line1.StripPreviewText(76);
                    customerAddress.Line2 = customerAddress.Line2.StripPreviewText(76);

                    customerContact.AddContactAddress(customerAddress);
                    customerContact.SaveChanges();
                }
            }
        }

        /// <summary>
        /// Processes the checkout details from PayPal, update changed addresses appropriately
        /// </summary>
        /// <param name="chkOutDetailsResponse">The CHK out details response.</param>
        /// <param name="cart">The cart.</param>
        /// <param name="billAddEmptyMsg">The bill add empty MSG.</param>
        /// <param name="shipAddEmptyMsg">The ship add empty MSG.</param>
        private void ProcessExChkOutDetails(GetExpressCheckoutDetailsResponseType chkOutDetailsResponse, Cart cart,
            ref string billAddEmptyMsg, ref string shipAddEmptyMsg, ref OrderAddress modifiedBillingAddress,
            ref OrderAddress modifiedShippingAddress, ref AddressType newShippingAddressFromPayPal)
        {
            GetExpressCheckoutDetailsResponseDetailsType details = chkOutDetailsResponse.GetExpressCheckoutDetailsResponseDetails;
            AddressType payPalBillAdd = details.BillingAddress;
            
            string cancelUrl = Utilities.GetUrlValueFromStartPage("ShoppingCartPage");

            //update billing address
            if (payPalBillAdd != null)
            {
                OrderAddress billAdd = new CartHelper(cart).FindAddressByName(cart.OrderForms[0].BillingAddressId);
                if (billAdd == null)
                {
                    // TODO: Utilities.UpdateCartInstanceId(cart);
                    ProcessUnsuccessfulTransaction(cancelUrl, LocalizationService.GetString("CommitTranErrorCartReset"));
                }
                if (string.IsNullOrEmpty(payPalBillAdd.Phone))
                {
                    if (details.PayerInfo != null && !string.IsNullOrEmpty(details.PayerInfo.ContactPhone))
                    {
                        payPalBillAdd.Phone = details.PayerInfo.ContactPhone;
                    }
                }
                if (IsAddressChanged(billAdd, payPalBillAdd))
                {
                    billAdd.City = payPalBillAdd.CityName;
                    billAdd.CountryCode = GetAlpha3CountryCode(payPalBillAdd.Country.ToString().ToUpperInvariant());
                    billAdd.DaytimePhoneNumber = payPalBillAdd.Phone;
                    billAdd.EveningPhoneNumber = payPalBillAdd.Phone;
                    billAdd.Line1 = payPalBillAdd.Street1;
                    billAdd.Line2 = payPalBillAdd.Street2;
                    billAdd.PostalCode = payPalBillAdd.PostalCode;
                    billAdd.State = GetStateName(payPalBillAdd.StateOrProvince);
                    billAdd.Email = details.PayerInfo.Payer;
                    string name = payPalBillAdd.Name.Trim();
                    billAdd.Name = name.StripPreviewText(46);
                    int index = name.IndexOf(' ');
                    billAdd.FirstName = index >= 0 ? name.Substring(0, index) : name;
                    billAdd.LastName = index >= 0 ? name.Substring(index + 1) : string.Empty;
                    modifiedBillingAddress = billAdd;

                    // Update billingAddressId when billing address was changed
                    cart.OrderForms[0].BillingAddressId = billAdd.Name;
                }
            }
            else
            {
                billAddEmptyMsg = Utilities.Translate("CommitTranErrorPayPalBillingAddressEmpty");
            }

            AddressType payPalShipAdd = details.PaymentDetails[0].ShipToAddress;
            // update shipping address
            if (payPalShipAdd != null)
            {
                var cartHelper = new CartHelper(cart);
                OrderAddress shipAdd = cartHelper.FindAddressByName(cart.OrderForms[0].Shipments[0].ShippingAddressId);
                if (shipAdd == null)
                {
                    // TODO: Utilities.UpdateCartInstanceId(cart);
                    ProcessUnsuccessfulTransaction(cancelUrl, Utilities.Translate("CommitTranErrorCartReset"));
                }
                if (string.IsNullOrEmpty(payPalShipAdd.Phone))
                {
                    if (details.PayerInfo != null && !string.IsNullOrEmpty(details.PayerInfo.ContactPhone))
                    {
                        payPalShipAdd.Phone = details.PayerInfo.ContactPhone;
                    }
                }
                if (IsAddressChanged(shipAdd, payPalShipAdd))
                {
                    string name = payPalShipAdd.Name.Trim().StripPreviewText(46);
                    // add new shipping address to OrderAddresses, so it can be found later, when getting shipping address.
                    // or update the existing one, in case the address was updated in PayPal
                    var existingShippingAddress = cartHelper.FindAddressByName(name);
                    if (existingShippingAddress == null)
                    {
                        shipAdd = new OrderAddress();
                    }
                    else
                    {
                        // keep the original order address untouch, otherwise it might update/remove the billing address
                        var addressEntity = new AddressEntity();
                        OrderAddress.CopyOrderAddressToCustomerAddress(existingShippingAddress, addressEntity);
                        shipAdd = new OrderAddress(addressEntity);
                    }

                    shipAdd.City = payPalShipAdd.CityName;
                    shipAdd.CountryCode = GetAlpha3CountryCode(payPalShipAdd.Country.ToString().ToUpperInvariant());
                    shipAdd.DaytimePhoneNumber = payPalShipAdd.Phone;
                    shipAdd.EveningPhoneNumber = payPalShipAdd.Phone;
                    shipAdd.Line1 = payPalShipAdd.Street1;
                    shipAdd.Line2 = payPalShipAdd.Street2;
                    shipAdd.PostalCode = payPalShipAdd.PostalCode;
                    shipAdd.State = GetStateName(payPalShipAdd.StateOrProvince);
                    shipAdd.Email = details.PayerInfo.Payer;
                    shipAdd.Name = name;
                    int index = name.IndexOf(' ');
                    shipAdd.FirstName = index >= 0 ? name.Substring(0, index) : name;
                    shipAdd.LastName = index >= 0 ? name.Substring(index + 1) : string.Empty;

                    // Update shippingAddressId when shipping address was changed
                    cart.OrderForms[0].Shipments[0].ShippingAddressId = shipAdd.Name;

                    if (existingShippingAddress == null)
                    {
                        modifiedShippingAddress = shipAdd;
                        cart.OrderAddresses.Add(shipAdd);
                    }
                }
                newShippingAddressFromPayPal = payPalShipAdd;
            }
            else
            {
                shipAddEmptyMsg = LocalizationService.GetString("CommitTranErrorPayPalShippingAddressEmpty");
            }
        }

        /// <summary>
        /// Determines whether the address is changed.
        /// </summary>
        /// <param name="orderAddress">The order address.</param>
        /// <param name="addressType">Type of the address.</param>
        /// <returns>
        /// 	<c>true</c> if [is address changed] [the specified order address]; otherwise, <c>false</c>.
        /// </returns>
        private bool IsAddressChanged(OrderAddress orderAddress, AddressType addressType)
        {
            if (!string.Equals(orderAddress.City, addressType.CityName, StringComparison.OrdinalIgnoreCase)
                || !string.Equals(orderAddress.CountryCode, GetAlpha3CountryCode(addressType.Country.ToString().ToUpperInvariant()), StringComparison.OrdinalIgnoreCase)
                || (!string.Equals(orderAddress.DaytimePhoneNumber ?? string.Empty, addressType.Phone ?? string.Empty, StringComparison.OrdinalIgnoreCase))
                || !string.Equals(orderAddress.Line1, addressType.Street1, StringComparison.OrdinalIgnoreCase)
                || (!string.Equals(orderAddress.Line2 ?? string.Empty, addressType.Street2 ?? string.Empty, StringComparison.OrdinalIgnoreCase))
                || !string.Equals(orderAddress.PostalCode, addressType.PostalCode, StringComparison.OrdinalIgnoreCase)
                || !string.Equals(orderAddress.State, GetStateName(addressType.StateOrProvince), StringComparison.OrdinalIgnoreCase)
                || !string.Equals(orderAddress.FirstName + ' ' + orderAddress.LastName, addressType.Name, StringComparison.OrdinalIgnoreCase))
            {
                return true;
            }
            return false;
        }

        /// <summary>
        /// Gets the MD5 key, in combination with this.HashKey
        /// </summary>
        /// <param name="hashString">The hash string.</param>
        /// <returns></returns>
        internal static string GetMD5Key(string hashString)
        {
            MD5CryptoServiceProvider md5Crypto = new MD5CryptoServiceProvider();
            byte[] arrBytes = System.Text.Encoding.UTF8.GetBytes(HashKey + hashString);
            arrBytes = md5Crypto.ComputeHash(arrBytes);
            StringBuilder sb = new StringBuilder();
            foreach (byte b in arrBytes)
            {
                sb.Append(b.ToString("x2").ToLower());
            }
            return sb.ToString();
        }


        /// <summary>
        /// Construct the Gets the PayPal payment details from our payment and Cart to pass onto PayPal.
        /// </summary>
        /// <remarks>
        /// <p>the PayPal payment detail can be a bit different from OrderForm because 
        /// sometimes cart total calculated by Commerce is different with cart total calculated by PalPal, 
        /// though this difference is very small (~0.01 currency).
        /// We adjust this into an addtional item to ensure PayPal shows the same total number with Commerce.
        /// </p>
        /// <p>
        /// We also add the Order discount (if any) as and special item with negative price to PayPal payment detail
        /// </p>
        /// <p>
        /// See detail about PayPal struct type in this link <seealso cref="https://cms.paypal.com/mx/cgi-bin/?cmd=_render-content&content_ID=developer/e_howto_api_soap_r_GetExpressCheckoutDetails#id0848IH0064Y__id099MB0BB0UX"/>
        /// </p>
        /// </remarks>
        /// <param name="payment">the payment to take info (Total, LineItem, ...) from</param>
        /// <param name="orderNumber">the orderNumber (to be InvoiceID to pass to PayPal)</param>
        /// <returns>the PayPal payment detail to pass to API request</returns>
        private PaymentDetailsType GetPaymentDetailsType(Mediachase.Commerce.Orders.Payment payment, string orderNumber)
        {
            var ppPayment = new PaymentDetailsType();

            ppPayment.ButtonSource = "Episerver_Cart_EC"; // (Optional) An identification code for use by third-party applications to identify transactions. Character length and limitations: 32 single-byte alphanumeric characters
            ppPayment.InvoiceID = orderNumber;  // Character length and limitations: 127 single-byte alphanumeric characters
            ppPayment.Custom = payment.Parent.Parent.CustomerId.ToString() + "|" + ppPayment.InvoiceID; // A free-form field for your own use. Character length and limitations: 256 single-byte alphanumeric characters
            // NOTE: ppPayment.OrderDescription = 127 single-byte alphanumeric characters string
            // NOTE: ppPayment.TransactionId = string, provided if you have transactionId in your Commerce system // (Optional) Transaction identification number of the transaction that was created.;

            // (Optional) Your URL for receiving Instant Payment Notification (IPN) about this transaction. If you do not specify this value in the request, the notification URL from your Merchant Profile is used, if one exists. 
            // IMPORTANT:The notify URL only applies to DoExpressCheckoutPayment. This value is ignored when set in SetExpressCheckout or GetExpressCheckoutDetails.
            // Character length and limitations: 2,048 single-byte alphanumeric characters
            ppPayment.NotifyURL = _notifyUrl;
            
            var currencyCode = GetCurrencyCode(payment);
            var currency = new Currency(currencyCode.ToString());
            decimal totalOrder = currency.Round(payment.Amount);
            decimal totalShipping = currency.Round(payment.Parent.ShippingTotal);
            decimal totalHandling = currency.Round(payment.Parent.HandlingTotal);
            decimal totalTax = currency.Round(payment.Parent.TaxTotal);
            
            decimal lineitemTotal = 0;

            var ppItemDetails = new List<PaymentDetailsItemType>();
            for (int i = 0; i < payment.Parent.LineItems.Count; i++)
            {
                var lineitem = payment.Parent.LineItems[i];
                
                var ppItemDetail = new PaymentDetailsItemType();
                ppItemDetail.Name = lineitem.DisplayName;
                ppItemDetail.Number = lineitem.Code;
                ppItemDetail.Description = lineitem.Description.StripPreviewText(100);  // Character length and limitations: 127 single-byte characters
                ppItemDetail.Quantity = Convert.ToInt32(lineitem.Quantity.ToString("0"));
                
                // recalculate final unit price after all kind of discounts are subtracted from item.ListPrice
                decimal finalUnitPrice = currency.Round(lineitem.ExtendedPrice / lineitem.Quantity); 
                ppItemDetail.Amount = finalUnitPrice.ToPayPalAmount(currencyCode);
                
                lineitemTotal += finalUnitPrice * lineitem.Quantity;
                ppItemDetails.Add(ppItemDetail);
            }

            #region Calculate the adjustment, if needed
            
            // this adjustment also include the giftcard (in sample)
            decimal orderAdjustment = totalOrder - totalShipping - totalHandling - totalTax - lineitemTotal;
            decimal adjustmentForShipping = 0;
            if (orderAdjustment != 0 || // adjustment for gift card/(order level) promotion case
                
                lineitemTotal == 0  // in this case, the promotion (or discount) make all lineItemTotal zero, but buyer still have to pay shipping (and/or handling, tax). 
                // We still need to adjust lineItemTotal for Paypal accepting (need to be greater than zero)
                )
            {
                var ppItemDetail = new PaymentDetailsItemType();
                ppItemDetail.Name = "Order adjustment";
                ppItemDetail.Number = "ORDERADJUSTMENT";
                ppItemDetail.Description = "GiftCard, Discount at OrderLevel and/or PayPal-Commerce-calculating difference in cart total";  // Character length and limitations: 127 single-byte characters
                ppItemDetail.Quantity = 1;

                var predictLineitemTotal = lineitemTotal + orderAdjustment;
                if (predictLineitemTotal <= 0)
                {
                    // can't overpaid for item. E.g.: total item amount is 68, orderAdjustment is -70, PayPal will refuse ItemTotal = -2
                    // we need to push -2 to shippingTotal or shippingDiscount

                    // HACK: Paypal will not accept an item total of $0, even if there is a shipping fee. The Item total must be at least 1 cent/penny. 
                    // We need to take 1 cent/penny from adjustmentForLineItemTotal and push to adjustmentForShipping
                    orderAdjustment = (-lineitemTotal + 0.01m); // -68 + 0.01 = -67.99
                    adjustmentForShipping = predictLineitemTotal - 0.01m;   // -2 - 0.01 = -2.01
                }
                else
                {
                    // this case means: due to PayPal calculation, buyer need to pay more that what Commerce calculate. Because:
                    // sometimes cart total calculated by Commerce is different with
                    // cart total calculated by PalPal, though this difference is very small (~0.01 currency)
                    // We adjust the items total to make up for that, to ensure PayPal shows the same total number with Commerce
                }
                ppItemDetail.Amount = orderAdjustment.ToPayPalAmount(currencyCode);
                lineitemTotal += orderAdjustment;   // re-adjust the lineItemTotal
                ppItemDetails.Add(ppItemDetail);
            }

            if (adjustmentForShipping > 0)
            {
                totalShipping += adjustmentForShipping;
            }
            else
            {
                // Shipping discount for this order. You specify this value as a negative number.
                // NOTE:Character length and limitations: Must not exceed $10,000 USD in any currency. 
                // No currency symbol. Regardless of currency, decimal separator must be a period (.), and the optional thousands separator must be a comma (,). 
                // Equivalent to nine characters maximum for USD. 
                // NOTE:You must set the currencyID attribute to one of the three-character currency codes for any of the supported PayPal currencies.
                ppPayment.ShippingDiscount = adjustmentForShipping.ToPayPalAmount(currencyCode); 
            }

            #endregion
            
            ppPayment.OrderTotal = totalOrder.ToPayPalAmount(currencyCode);
            ppPayment.ShippingTotal = totalShipping.ToPayPalAmount(currencyCode);
            ppPayment.HandlingTotal = totalHandling.ToPayPalAmount(currencyCode);
            ppPayment.TaxTotal = totalTax.ToPayPalAmount(currencyCode);
            ppPayment.ItemTotal = lineitemTotal.ToPayPalAmount(currencyCode);
            
            ppPayment.PaymentDetailsItem = ppItemDetails; 
            ppPayment.ShipToAddress = GetShippingAddress(payment);

            if (payment.Parent.Shipments != null && payment.Parent.Shipments.Count > 1) 
            {
                // (Optional) The value 1 indicates that this payment is associated with multiple shipping addresses. Character length and limitations: Four single-byte numeric characters.
                ppPayment.MultiShipping = "1";
            }
            
            return ppPayment;
        }

        /// <summary>
        /// Generates the order number by "PO" + cart.OrderGroupId + Random(100-999)
        /// </summary>
        /// <param name="cart">The OrderGroupId, used to take from from Cart.OrderGroupId</param>
        /// <returns>Random string to be OrderNumber</returns>
        private string GenerateOrderNumber(int OrderGroupId)
        {
            string str = new Random().Next(100, 999).ToString();
            return string.Format("PO{0}{1}", OrderGroupId, str);
        }

        /// <summary>
        /// Gets the billing address.
        /// </summary>
        /// <param name="payment">The payment.</param>
        /// <returns></returns>
        /// <value>The billing address.</value>
        public AddressType GetBillingAddress(Mediachase.Commerce.Orders.Payment payment)
        {
            AddressType billAdd = new AddressType();
            CartHelper helper = new CartHelper(Cart.DefaultName);
            _address = helper.FindAddressByName(payment.Parent.BillingAddressId);
            billAdd.CityName = _address.City;
            billAdd.Country = GetAlpha2CountryCode(_address);
            billAdd.Street1 = _address.Line1;
            billAdd.Street2 = _address.Line2;
            billAdd.PostalCode = _address.PostalCode;
            billAdd.StateOrProvince = GetStateCode(_address.State);
            billAdd.Phone = (string.IsNullOrEmpty(_address.DaytimePhoneNumber) ? _address.EveningPhoneNumber : _address.DaytimePhoneNumber);
            billAdd.Name = _address.FirstName + " " + _address.LastName;
            return billAdd;
        }

        /// <summary>
        /// Gets the shipping address.
        /// </summary>
        /// <value>The shipping address.</value>
        public AddressType GetShippingAddress(Mediachase.Commerce.Orders.Payment payment)
        {
            AddressType shipingAdd = new AddressType();
            CartHelper helper = new CartHelper(payment.Parent.Parent as Cart);
            var shippingAddress = helper.FindAddressByName(payment.Parent.Shipments[0].ShippingAddressId);
            shipingAdd.CityName = shippingAddress.City;
            shipingAdd.Country = GetAlpha2CountryCode(shippingAddress);
            shipingAdd.Street1 = shippingAddress.Line1;
            shipingAdd.Street2 = shippingAddress.Line2;
            shipingAdd.PostalCode = shippingAddress.PostalCode;
            shipingAdd.StateOrProvince = GetStateCode(shippingAddress.State);
            shipingAdd.Phone = (string.IsNullOrEmpty(shippingAddress.DaytimePhoneNumber) ? shippingAddress.EveningPhoneNumber : shippingAddress.DaytimePhoneNumber);
            shipingAdd.Name = shippingAddress.FirstName + " " + shippingAddress.LastName;
            return shipingAdd;
        }


        #region For redirection inside iframe

        /// <summary>
        /// Redirects to PayPal.com to process payment.
        /// </summary>
        /// <param name="urlNeedToRedirectByParent"></param>
        private void RedirectParentPage(string urlNeedToRedirectByParent)
        {
            HttpContext.Current.Response.Redirect(urlNeedToRedirectByParent, true);
        }

        #endregion

        private static LocalizationService LocalizationService
        {
            get { return ServiceLocator.Current.GetInstance<LocalizationService>(); }
        }
    }
}
